#ifndef __snf3x3_H__
#define __snf3x3_H__

#ifdef __cplusplus
extern "C" {
#endif

#define SNF3X3_MAJOR_VERSION 0
#define SNF3X3_MINOR_VERSION 1
#define SNF3X3_MICRO_VERSION 0

int snf3x3(long A[3][3], long P[3][3], long Q[3][3]);

#ifdef __cplusplus
}
#endif
#endif
